from numbers import Number
from pdf_scout.extract import extract_all_words
from pdf_scout.scoring import score_words
from PyPDF2 import PdfMerger
from rich import print as rprint
from time import time
from typing import List, Tuple, TypedDict
from typing import Optional
import pdfplumber
import typer


class Bookmark(TypedDict):
    title: str
    page_number: str
    scroll_distance: Number


def write_bookmarks(
    input_path: str, output_path: str, bookmarks: List[Bookmark]
) -> None:
    merger = PdfMerger()
    merger.append(input_path, import_outline=False)  # disregard existing outline

    parent_bookmarks: List[Tuple(int, any)] = []
    # last item in list is last outline item added

    add_bookmark_to_writer = lambda writer, bookmark, parent: writer.add_outline_item(
        bookmark["title"],
        bookmark["page_number"] - 1,
        parent,
        None,
        False,
        False,
        "/FitH",
        bookmark["scroll_distance"],
    )
    get_last_bookmark = (
        lambda parent_bs: parent_bs[-1][1] if len(parent_bs) >= 1 else None
    )

    for rank, bookmark in bookmarks:
        add_bookmark = lambda p: add_bookmark_to_writer(merger, bookmark, p)
        # rprint(rank, bookmark["title"])

        if len(parent_bookmarks) == 0:
            new_bookmark = add_bookmark(None)
            parent_bookmarks.append((rank, new_bookmark))
        else:
            last = parent_bookmarks[-1]
            last_rank, last_bookmark = last
            if last_rank < rank:
                new_bookmark = add_bookmark(last_bookmark)
                parent_bookmarks.append((rank, new_bookmark))
            elif last_rank == rank:
                parent_bookmarks.pop()
                parent_bookmark = get_last_bookmark(parent_bookmarks)
                new_bookmark = add_bookmark(parent_bookmark)
                parent_bookmarks.append((rank, new_bookmark))
            elif last_rank > rank:
                parent_bookmarks.pop()
                if len(parent_bookmarks) >= 1:
                    parent_bookmarks.pop()
                parent_bookmark = get_last_bookmark(parent_bookmarks)
                new_bookmark = add_bookmark(parent_bookmark)

    merger.write(output_path)
    merger.close()
    return None


def open_pdf_file(input_path: str) -> pdfplumber.PDF:
    return pdfplumber.open(input_path)


def get_words_from_pdf_file(pdf_file: pdfplumber.PDF):
    all_words = extract_all_words(pdf_file)
    return all_words


def add_bookmarks_to_pdf(input_path: str, output_path: str = "", levels=3):
    if len(output_path) == 0:
        input_path_start, _ = input_path.split(".pdf")
        output_path = f"{input_path_start}-out.pdf"

    pdf_file = open_pdf_file(input_path)
    all_words = get_words_from_pdf_file(pdf_file)
    scored_words = score_words(all_words)

    top_scores: List[Number] = sorted(
        list(set([score["overall"] for score, _ in scored_words])), reverse=True
    )[0:levels]
    top_scored_words = [
        [top_scores.index(score["overall"]), word]
        for score, word in scored_words
        if score["overall"] in top_scores
    ]

    bookmarks: List[Tuple[int, Bookmark]] = [
        (
            rank,
            dict(
                title=word["text"],
                page_number=word["page_number"],
                scroll_distance=(
                    pdf_file.pages[word["page_number"] - 1].height
                    - word["top"]
                    + word["bottom"]
                    - word["top"]
                ),
            ),
        )
        for rank, word in top_scored_words
    ]

    # rprint(locals())

    write_bookmarks(input_path, output_path, bookmarks)

    pdf_file.close()


def main(input_file_path: str, output_file_path: Optional[str] = typer.Argument("")):
    start_time = time()
    if input_file_path is None or len(input_file_path) == 0:
        print("Error: file_path not provided")
        raise typer.Exit(code=1)
    add_bookmarks_to_pdf(input_file_path, output_file_path)
    end_time = time()
    rprint(f"Finished in {end_time - start_time}s")


def start():
    typer.run(main)


if __name__ == "__main__":
    start()



from numbers import Number
from rich import print as rprint
from typing import List, Tuple, TypedDict
import re
import statistics


def score_font_name(font_name: str) -> Number:
    if re.search(r"Bold$", font_name):
        return 20
    elif re.search(r"Semibold", font_name):
        return 15
    elif re.search(r"BoldItalic$", font_name):
        return 10
    if re.search(r"(It|Italic|Oblique)$", font_name):
        return 5
    return 0


def score_font_size(font_size: Number) -> Number:
    # TODO: make more sophisticated; steps should be exponential
    return font_size


def score_word_length(length: int) -> Number:
    MIN_THRESHOLD = 4  # penalise if <= this length
    MAX_THRESHOLD = 80  # don't penalise if <= this length
    STARTING_SCORE = 100
    if length >= MIN_THRESHOLD and length <= MAX_THRESHOLD:
        return STARTING_SCORE
    elif length > MAX_THRESHOLD:
        return STARTING_SCORE - (length if length <= STARTING_SCORE else STARTING_SCORE)
    elif length < 4:
        return STARTING_SCORE * length / MIN_THRESHOLD


class HeadingScore(TypedDict):
    font_name: Number
    font_size: Number
    word_length: Number
    font: Number
    overall: Number


def get_heading_score(word) -> HeadingScore:
    font_name: str = word["fontname"]
    font_size: Number = word["size"]
    length: int = len(word["text"].strip(" []()./|"))

    font_name_score = score_font_name(font_name)
    font_size_score = score_font_size(font_size)
    word_length_score = score_word_length(length)

    score = round(font_name_score + font_size_score + word_length_score, 2)
    return {
        "font_name": round(font_name_score, 2),
        "font_size": round(font_size_score, 2),
        "word_length": round(word_length_score, 2),
        "font": round(font_name_score + font_size_score, 2),
        "overall": score,
    }


def guess_body_score(word_list: Tuple[HeadingScore, any]) -> Number:
    return statistics.mode([score["font"] for score, _ in word_list])


def score_words(all_words: List[any]):
    scored_words: List[Tuple[HeadingScore, any]] = [
        (get_heading_score(word), word) for word in all_words
    ]
    body_score = guess_body_score(scored_words)
    # ignore all body text
    scored_words = [
        (score, word) for score, word in scored_words if score["font"] != body_score
    ]

    # rprint(locals())

    return scored_words



from itertools import groupby
from numbers import Number
from operator import itemgetter
from rich import print as rprint
from typing import List, Tuple
import statistics


def guess_left_margin(words) -> List[Number]:
    words_x0 = [round(word["x0"]) for word in words]
    counts = [(x0, words_x0.count(x0)) for x0 in set(words_x0)]
    threshold_counts = [
        (left_margin, count)
        for left_margin, count in counts
        if count >= 0.1 * len(words_x0)
    ]

    # rprint(locals())

    if len(threshold_counts) == 2:
        # assume different left margin on odd and even pages
        return [left_margin for left_margin, _ in threshold_counts]
    else:
        # assume uniform left margin
        return (
            [threshold_counts[0][0]]
            if len(threshold_counts) == 1
            else [max(counts, key=lambda x: x[1])[0]]
        )


def add_line_spacing_to_words(pdf_file, all_words):
    # add distance from previous & next line
    # assumes all lines are perfectly horizontal and of full width
    line_positions: List[Tuple[int, List[Number]]] = [
        (
            page_number,
            sorted(list(set([get_word_line_position(word) for word in words]))),
        )
        for page_number, words in groupby(all_words, key=itemgetter("page_number"))
    ]
    return [
        add_line_spacing_to_word(line_positions, word, pdf_file) for word in all_words
    ]


def guess_body_spacing(words) -> Tuple[Number, Number]:
    return (
        statistics.mode([word["top_spacing"] for word in words]),
        statistics.mode([word["bottom_spacing"] for word in words]),
    )


def guess_body_font_size(words) -> Number:
    return statistics.mode([word["size"] for word in words])


def get_word_line_position(word) -> Number:
    return word["top"]


def add_line_spacing_to_word(
    line_positions: List[Tuple[int, List[Number]]], word, pdf_file
):
    page_line_positions = [
        page_lines
        for page_number, page_lines in line_positions
        if page_number == word["page_number"]
    ][0]
    index = page_line_positions.index(get_word_line_position(word))
    cur_line_position = page_line_positions[index]
    prev_line_position = page_line_positions[index - 1] if index != 0 else 0
    next_line_position = (
        page_line_positions[index + 1]
        if index < len(page_line_positions) - 1
        else pdf_file.pages[word["page_number"] - 1].height
    )
    return dict(
        **word,
        top_spacing=round(cur_line_position - prev_line_position, 2),
        bottom_spacing=round(next_line_position - cur_line_position, 2),
    )


def raw_extract_words(pdf_file) -> List[dict[str, any]]:
    all_words = [
        word
        for page_list in (
            [
                dict(**word, page_number=page.page_number)
                for word in page.extract_words(
                    keep_blank_chars=True,
                    use_text_flow=True,
                    extra_attrs=["fontname", "size"],
                )
            ]
            for page in pdf_file.pages
        )
        for word in page_list
    ]
    return all_words


def extract_all_words(pdf_file) -> List[dict[str, any]]:
    all_words = raw_extract_words(pdf_file)
    body_font_size = guess_body_font_size(all_words)
    all_words = add_line_spacing_to_words(pdf_file, all_words)

    body_top_spacing, body_bottom_spacing = guess_body_spacing(all_words)

    # TODO: add some margin of appreciation to account for indented headers, footnotes, etc
    # TODO: handle center-aligned text
    left_margins = guess_left_margin(all_words)

    all_words = [
        word
        for word in all_words
        if (
            (
                # ignore all words with normal paragraph spacing
                # 5% because line spacing is not always precise
                word["top_spacing"] >= body_top_spacing * 1.05
                and word["bottom_spacing"] >= body_bottom_spacing * 1.05
            )
            and (  # ignore all text smaller than body font size
                word["size"] >= body_font_size
            )
            and (  # ignore all words not at left margin
                round(word["x0"]) in left_margins
            )
        )
    ]

    # rprint(locals())

    return all_words



from joblib import Memory

CACHE_DIR = "./.cache"
memory = Memory(CACHE_DIR, verbose=0)


def cache(input_function):
    """
    Only works with pure functions
    """
    return memory.cache(input_function)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pdf_scout']

package_data = \
{'': ['*']}

install_requires = \
['PyPDF2>=2.10.9,<3.0.0',
 'joblib>=1.2.0,<2.0.0',
 'pdfplumber>=0.7.4,<0.8.0',
 'rich>=12.5.1,<13.0.0',
 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['pdf_scout = pdf_scout.app:start']}

setup_kwargs = {
    'name': 'pdf-scout',
    'version': '0.0.3',
    'description': 'automatically create bookmarks in a PDF file',
    'long_description': "# pdf_scout\n\nThis CLI tool automatically generates PDF bookmarks (also known as an 'outline' or a 'table of contents') for computer-generated PDF documents.\n\nYou can install it globally via pip:\n\n```\npip install pdf_scout\npdf_scout ./my_document.pdf\npip uninstall pdf_scout\n```\n\n![screenshot](./assets/screenshot.png)\n\nThis project is a work in progress and will likely only generate accurate bookmarks for documents that conform to the following requirements:\n\n* Single column of text (not multiple columns)\n* Font size of header text >= font size of body text\n* Header text is justified or left-aligned\n\n## Development\n\nThis project manages its dependencies using [poetry](https://python-poetry.org) and is only supported for Python ^3.9. After installing poetry and entering the project folder, run the following to install the dependencies:\n\n```bash\npoetry install\n```\n\nTo open a virtualenv in the project folder with the dependencies, run:\n\n```bash\npoetry shell\n```\n\nTo run a script directly, run:\n\n```bash\npoetry run python ./src/app.py\n```\n\n### Tests\n\nThere are snapshot tests. Input PDFs are not provided at the moment, so you will have to populate the `/pdf` folder manually:\n\n```bash\npoetry run pytest\npoetry run pytest --snapshot-update\n```",
    'author': 'Huey',
    'author_email': 'hello@huey.xyz',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/hueyy/pdf_scout',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
